# Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from typing import Tuple, Type, Optional

import cuda.bindings.driver as cuda
import cutlass.utils as utils
import cutlass.pipeline as pipeline
from cutlass.cute.nvgpu import cpasync, tcgen05
import cutlass.utils.blackwell_helpers as sm100_utils
import cutlass
import cutlass.cute as cute
from cutlass.cute.typing import Int32, Float32, Int64

from cutlass._mlir.dialects import cute_nvgpu

"""
A NSA(Native Sparse Attention) Top-K Reduction Forward Pass for NVIDIA Blackwell SM100 architecture using Cute DSL.
"""


class FineGrainedReductionQK:
    def __init__(
        self,
        element_dtype: Type[cutlass.Numeric],
        acc_dtype: Type[cutlass.Numeric],
        k_value: int,
        selection_block_size: int,
        compress_block_sliding_stride: int,
        mma_tiler: Tuple[int, int, int],
        is_causal: bool = False,
    ):
        self.element_dtype = element_dtype
        self.acc_dtype = acc_dtype
        self.indices_dtype = Int32
        self.k_value = k_value
        self.selection_block_size = selection_block_size
        self.compress_block_sliding_stride = compress_block_sliding_stride
        self.num_elem_for_reduction = selection_block_size // compress_block_sliding_stride

        self.cluster_shape_mn = (1, 1)
        self.mma_tiler = mma_tiler
        self.is_causal = is_causal

        self.compute_warp_id = (0, 1, 2, 3)
        # self.reduce_warp_id = (4, 5, 6, 7)
        self.load_warp_id = 4
        self.mma_warp_id = 5
        self.epi_warp_id = 6
        self.threads_per_warp = 32
        self.num_compute_warps = len(self.compute_warp_id)
        # self.num_reduce_warps = len(self.reduce_warp_id)

        self.threads_per_cta = self.threads_per_warp * (self.num_compute_warps + 4)

        self.cta_sync_bar_id = 0
        self.tmem_alloc_sync_bar_id = 1
        self.compute_sync_bar_id = 2
        # self.reduce_sync_bar_id = 3
        # self.num_regs_compute = 240
        # self.num_regs_reduce = 96
        # self.num_regs_other = 72
        self.num_regs_compute = 256
        # self.num_regs_reduce = 96
        self.num_regs_other = 96

    @cute.jit
    def __call__(
        self,
        problem_size: Tuple[Int32, Int32, Int32, Int32, Int32, Int32],
        Q: cute.Tensor,
        K: cute.Tensor,
        LSE: cute.Tensor,
        Topk_scores: cute.Tensor,
        Topk_indices: cute.Tensor,
        softmax_scale_log2_e: Float32,
        cumulative_s_q: Optional[cute.Tensor],
        cumulative_s_k: Optional[cute.Tensor],
        stream: cuda.CUstream,
    ):
        b, s_q_max, s_k_max, _, _, _ = problem_size
        s_q, s_k = Q.shape[2], K.shape[2]
        h_q, h_k = Q.shape[1], K.shape[1]
        head_dim = Q.shape[3]
        h_r = h_q // h_k

        stride_b_q = s_q * head_dim * h_k * h_r if cumulative_s_q is None else 0
        stride_b_k = s_k * head_dim * h_k if cumulative_s_k is None else 0
        stride_b_lse = s_q * h_r * h_k if cumulative_s_q is None else 0
        stride_b_out = s_q * (s_k_max // self.num_elem_for_reduction) * h_k if cumulative_s_q is None else 0
        stride_b_topk_scores = s_q * self.k_value * h_k if cumulative_s_q is None else 0
        stride_b_topk_indices = s_q * self.k_value * h_k if cumulative_s_q is None else 0

        Q = cute.make_tensor(
            Q.iterator,
            cute.make_layout(
                (s_q, head_dim, (h_r, h_k, b)),
                stride=(
                    head_dim * h_r * h_k,
                    1,
                    (
                        head_dim * h_k,
                        head_dim,
                        stride_b_q,
                    ),  # TODO: head ordering is diff?
                ),
            ),
        )

        K = cute.make_tensor(
            K.iterator,
            cute.make_layout(
                (s_k, head_dim, (1, h_k, b)),
                stride=(
                    head_dim * h_k,
                    1,
                    (0, head_dim, stride_b_k),
                ),
            ),
        )

        # reshape LSE to (s_q, h_r, h_k, b)
        LSE = cute.make_tensor(
            LSE.iterator,
            cute.make_layout(
                (s_q, h_r, h_k, b),
                stride=(1, s_q * h_k, s_q, stride_b_lse),
            ),
        )

        Topk_scores = cute.make_tensor(
            Topk_scores.iterator,
            cute.make_layout(
                (s_q, self.k_value, (1, h_k, b)),
                stride=(self.k_value * h_k, 1, (0, self.k_value, stride_b_topk_scores)),
            ),
        )

        Topk_indices = cute.make_tensor(
            Topk_indices.iterator,
            cute.make_layout(
                (s_q, self.k_value, (1, h_k, b)),  # (s_q, k, h_k, b)
                stride=(
                    self.k_value * h_k,
                    1,
                    (0, self.k_value, stride_b_topk_indices),
                ),
            ),
        )

        self.q_major_mode = utils.LayoutEnum.from_tensor(Q).mma_major_mode()
        self.k_major_mode = utils.LayoutEnum.from_tensor(K).mma_major_mode()
        self.load_mma_Q_stage = 2
        self.load_mma_K_stage = 1
        self.load_compute_LSE_stage = 1
        self.mma_compute_S_stage = 1
        cta_group = tcgen05.CtaGroup.ONE

        QK_tiled_mma = sm100_utils.make_trivial_tiled_mma(
            self.element_dtype,
            self.element_dtype,
            self.q_major_mode,
            self.k_major_mode,
            self.acc_dtype,
            cta_group,
            self.mma_tiler[:2],
        )
        Q_smem_layout_staged = sm100_utils.make_smem_layout_a(
            QK_tiled_mma,
            self.mma_tiler,
            self.element_dtype,
            self.load_mma_Q_stage,
        )
        K_smem_layout_staged = sm100_utils.make_smem_layout_b(
            QK_tiled_mma,
            self.mma_tiler,
            self.element_dtype,
            self.load_mma_K_stage,
        )
        LSE_smem_layout = cute.make_layout((self.mma_tiler[0], 1))

        self.epi_tile = (self.mma_tiler[0], self.k_value)
        self.cluster_shape_mnk = (*self.cluster_shape_mn, 1)
        self.cluster_layout_vmnk = cute.tiled_divide(
            cute.make_layout(self.cluster_shape_mnk),
            (QK_tiled_mma.thr_id.shape,),
        )

        tma_load_op = cute.nvgpu.cpasync.CopyBulkTensorTileG2SOp(cta_group)
        tma_store_op = cute.nvgpu.cpasync.CopyBulkTensorTileS2GOp()

        Q_smem_layout = cute.select(Q_smem_layout_staged, mode=[0, 1, 2])
        K_smem_layout = cute.select(K_smem_layout_staged, mode=[0, 1, 2])

        tma_atom_Q, tma_tensor_Q = cute.nvgpu.make_tiled_tma_atom_A(
            tma_load_op,
            Q,
            Q_smem_layout,
            self.mma_tiler,
            QK_tiled_mma,
            self.cluster_layout_vmnk,
        )
        tma_atom_K, tma_tensor_K = cute.nvgpu.make_tiled_tma_atom_B(
            tma_load_op,
            K,
            K_smem_layout,
            self.mma_tiler,
            QK_tiled_mma,
            self.cluster_layout_vmnk,
        )

        self.tma_copy_Q_bytes = cute.size_in_bytes(self.element_dtype, Q_smem_layout)
        self.tma_copy_K_bytes = cute.size_in_bytes(self.element_dtype, K_smem_layout)

        @cute.struct
        class SharedStorage:
            load_mma_Q_mbar_ptr: cute.struct.MemRange[cutlass.Int64, self.load_mma_Q_stage * 2]
            load_mma_K_mbar_ptr: cute.struct.MemRange[cutlass.Int64, self.load_mma_K_stage * 2]
            load_compute_LSE_mbar_ptr: cute.struct.MemRange[cutlass.Int64, self.load_compute_LSE_stage * 2]
            mma_compute_S_mbar_ptr: cute.struct.MemRange[cutlass.Int64, self.mma_compute_S_stage * 2]

            sQ: cute.struct.Align[
                cute.struct.MemRange[self.element_dtype, cute.cosize(Q_smem_layout_staged)],
                1024,
            ]
            sK: cute.struct.Align[
                cute.struct.MemRange[self.element_dtype, cute.cosize(K_smem_layout_staged)],
                1024,
            ]
            sLSE: cute.struct.Align[
                cute.struct.MemRange[self.acc_dtype, cute.cosize(LSE_smem_layout)],
                1024,
            ]

            tmem_holding_buf: Int32

        self.shared_storage = SharedStorage

        grid_shape = (cute.ceil_div(s_q_max, self.mma_tiler[0]), h_k, b)
        block_shape = (self.threads_per_cta, 1, 1)

        self.kernel(
            problem_size,
            QK_tiled_mma,
            tma_atom_Q,
            tma_tensor_Q,
            tma_atom_K,
            tma_tensor_K,
            LSE,
            cumulative_s_q,
            cumulative_s_k,
            Topk_scores,
            Topk_indices,
            Q_smem_layout_staged,
            K_smem_layout_staged,
            LSE_smem_layout,
            softmax_scale_log2_e,
        ).launch(
            grid=grid_shape,
            block=block_shape,
            smem=SharedStorage.size_in_bytes(),
            stream=stream,
            min_blocks_per_mp=1,
        )

    def make_and_init_load_mma_Q_pipeline(self, load_mma_Q_mbar_ptr):
        load_mma_Q_producer_group = pipeline.CooperativeGroup(pipeline.Agent.Thread, len([self.load_warp_id]))
        load_mma_Q_consumer_group = pipeline.CooperativeGroup(pipeline.Agent.Thread, len([self.mma_warp_id]))
        return pipeline.PipelineTmaUmma.create(
            barrier_storage=load_mma_Q_mbar_ptr,
            num_stages=self.load_mma_Q_stage,
            producer_group=load_mma_Q_producer_group,
            consumer_group=load_mma_Q_consumer_group,
            tx_count=self.tma_copy_Q_bytes,
        )

    def make_and_init_load_mma_K_pipeline(self, load_mma_K_mbar_ptr):
        load_mma_K_producer_group = pipeline.CooperativeGroup(pipeline.Agent.Thread, len([self.load_warp_id]))
        load_mma_K_consumer_group = pipeline.CooperativeGroup(pipeline.Agent.Thread, len([self.mma_warp_id]))
        return pipeline.PipelineTmaUmma.create(
            barrier_storage=load_mma_K_mbar_ptr,
            num_stages=self.load_mma_K_stage,
            producer_group=load_mma_K_producer_group,
            consumer_group=load_mma_K_consumer_group,
            tx_count=self.tma_copy_K_bytes,
        )

    def make_and_init_load_compute_LSE_pipeline(self, load_compute_lse_mbar_ptr):
        load_compute_lse_producer_group = pipeline.CooperativeGroup(
            pipeline.Agent.Thread,
            self.threads_per_warp,
        )
        load_compute_lse_consumer_group = pipeline.CooperativeGroup(
            pipeline.Agent.Thread,
            self.threads_per_warp * self.num_compute_warps,
        )
        return pipeline.PipelineCpAsync.create(
            barrier_storage=load_compute_lse_mbar_ptr,
            num_stages=self.load_compute_LSE_stage,
            producer_group=load_compute_lse_producer_group,
            consumer_group=load_compute_lse_consumer_group,
        )

    def make_and_init_mma_compute_S_pipeline(self, mma_compute_S_mbar_ptr):
        mma_compute_S_producer_group = pipeline.CooperativeGroup(
            pipeline.Agent.Thread,
            len([self.mma_warp_id]),
        )
        mma_compute_S_consumer_group = pipeline.CooperativeGroup(
            pipeline.Agent.Thread,
            self.num_compute_warps * self.threads_per_warp,
        )
        return pipeline.PipelineUmmaAsync.create(
            barrier_storage=mma_compute_S_mbar_ptr,
            num_stages=self.mma_compute_S_stage,
            producer_group=mma_compute_S_producer_group,
            consumer_group=mma_compute_S_consumer_group,
        )

    @cute.kernel
    def kernel(
        self,
        problem_size: Tuple[Int32, Int32, Int32, Int32, Int32, Int32],
        QK_tiled_mma: cute.TiledMma,
        tma_atom_Q: cute.CopyAtom,
        tma_tensor_Q: cute.Tensor,
        tma_atom_K: cute.CopyAtom,
        tma_tensor_K: cute.Tensor,
        LSE: cute.Tensor,
        cumulative_s_q: Optional[cute.Tensor],
        cumulative_s_k: Optional[cute.Tensor],
        Topk_scores: cute.Tensor,
        Topk_indices: cute.Tensor,
        Q_smem_layout_staged: cute.ComposedLayout,
        K_smem_layout_staged: cute.ComposedLayout,
        LSE_smem_layout: cute.Layout,
        softmax_scale_log2_e: Float32,
    ):
        tidx, tidy, _ = cute.arch.thread_idx()
        bidx, bidy, bidz = cute.arch.block_idx()

        warp_idx = cute.arch.warp_idx()
        warp_idx = cute.arch.make_warp_uniform(warp_idx)

        b, s_q_max, s_k_max, h_q, h_k, head_dim = problem_size
        h_r = h_q // h_k

        smem = utils.SmemAllocator()
        storage = smem.allocate(self.shared_storage)

        load_mma_Q_pipeline = self.make_and_init_load_mma_Q_pipeline(storage.load_mma_Q_mbar_ptr.data_ptr())
        load_mma_K_pipeline = self.make_and_init_load_mma_K_pipeline(storage.load_mma_K_mbar_ptr.data_ptr())
        load_compute_LSE_pipeline = self.make_and_init_load_compute_LSE_pipeline(storage.load_compute_LSE_mbar_ptr.data_ptr())
        mma_compute_S_pipeline = self.make_and_init_mma_compute_S_pipeline(storage.mma_compute_S_mbar_ptr.data_ptr())

        cute.arch.barrier(barrier_id=self.cta_sync_bar_id, number_of_threads=self.threads_per_cta)

        sQ = storage.sQ.get_tensor(Q_smem_layout_staged.outer, swizzle=Q_smem_layout_staged.inner)
        sK = storage.sK.get_tensor(K_smem_layout_staged.outer, swizzle=K_smem_layout_staged.inner)
        sLSE = storage.sLSE.get_tensor(LSE_smem_layout)

        block_offset = (Int32(0), Int32(0), Int32(0), (Int32(0), Int32(0), Int32(0)))
        cur_s_q = s_q_max
        cur_s_k = s_k_max
        if cutlass.const_expr(cumulative_s_q is not None):
            cur_s_q = cumulative_s_q[bidz + 1] - cumulative_s_q[bidz]
            cur_s_k = cumulative_s_k[bidz + 1] - cumulative_s_k[bidz]
            block_offset = (
                cumulative_s_q[bidz],
                cumulative_s_k[bidz],
                Int32(0),
                (Int32(0), Int32(0), Int32(0)),
            )

        k_tile_count = cute.ceil_div(cur_s_k, self.mma_tiler[1])
        k_tile_idx = 0
        if cutlass.const_expr(self.is_causal):
            q_max = min(cur_s_q, (bidx + 1) * self.mma_tiler[0])
            k_cols = min(cur_s_k, q_max // self.compress_block_sliding_stride)
            k_tile_count = cute.ceil_div(k_cols, self.mma_tiler[1])

        mQ = cute.domain_offset(cute.select(block_offset, mode=[0, 2, 3]), tma_tensor_Q)
        mK = cute.domain_offset(cute.select(block_offset, mode=[1, 2, 3]), tma_tensor_K)
        mTopk_scores = cute.make_tensor(
            Topk_scores.iterator + cute.assume(block_offset[0] * Topk_scores.stride[0], divby=self.k_value),
            Topk_scores.layout,
        )
        mTopk_indices = cute.make_tensor(
            Topk_indices.iterator + cute.assume(block_offset[0] * Topk_indices.stride[0], divby=self.k_value),
            Topk_indices.layout,
        )

        # (MMA_M, MMA_K, REST_M, REST_K, (H_r, H_k, B))
        gQ = cute.local_tile(mQ, cute.select(self.mma_tiler, mode=[0, 2]), (None, None, None))
        # (MMA_N, MMA_K, REST_N, REST_K, (1, H_k, B))
        gK = cute.local_tile(mK, cute.select(self.mma_tiler, mode=[0, 2]), (None, None, None))

        # (MMA_M, MMA_K, H_r)
        gQ = gQ[None, None, bidx, 0, (None, bidy, bidz)]
        # (MMA_N, MMA_K, REST_N)
        gK = gK[None, None, None, 0, (0, bidy, bidz)]

        thr_mma = QK_tiled_mma.get_slice(0)

        tSgQ = thr_mma.partition_A(gQ)
        tSgK = thr_mma.partition_B(gK)

        # tQgQ: (MMA, H_r)
        # tQsQ: (MMA, PIPE)
        tQsQ, tQgQ = cute.nvgpu.cpasync.tma_partition(
            tma_atom_Q,
            0,
            cute.make_layout(1),
            cute.group_modes(sQ, 0, 3),
            cute.group_modes(tSgQ, 0, 3),
        )

        # tKgK: (MMA, REST_N)
        # tKsK: (MMA, PIPE)
        tKsK, tKgK = cute.nvgpu.cpasync.tma_partition(
            tma_atom_K,
            0,
            cute.make_layout(1),
            cute.group_modes(sK, 0, 3),
            cute.group_modes(tSgK, 0, 3),
        )

        tSrQ = QK_tiled_mma.make_fragment_A(sQ)
        tSrK = QK_tiled_mma.make_fragment_B(sK)
        tStS_shape = QK_tiled_mma.partition_shape_C(cute.select(self.mma_tiler, mode=[0, 1]))
        # ((MMA_M, MMA_N), REST_M, REST_N)
        tStS = QK_tiled_mma.make_fragment_C(tStS_shape)
        # another tmem for reduction
        tStS_reduce = cute.make_tensor(tStS.iterator + self.mma_tiler[1], tStS.layout)

        if bidx * self.mma_tiler[0] < cur_s_q and k_tile_count > 0:

            load_iter_count = k_tile_count
            load_iter_index = k_tile_idx
            mma_iter_count = k_tile_count
            mma_iter_index = k_tile_idx
            compute_iter_count = k_tile_count
            compute_iter_index = k_tile_idx

            # LOAD Q K WARP
            if warp_idx == self.load_warp_id:
                # TODO: reconfig regs
                cute.arch.setmaxregister_decrease(self.num_regs_other)

                load_mma_Q_producer_state = pipeline.make_pipeline_state(pipeline.PipelineUserType.Producer, self.load_mma_Q_stage)
                load_mma_K_producer_state = pipeline.make_pipeline_state(pipeline.PipelineUserType.Producer, self.load_mma_K_stage)
                load_compute_LSE_producer_state = pipeline.make_pipeline_state(pipeline.PipelineUserType.Producer, self.load_compute_LSE_stage)

                while load_iter_count > 0:
                    # Wait for K to be empty
                    load_mma_K_pipeline.producer_acquire(load_mma_K_producer_state)
                    K_tma_barrier = load_mma_K_pipeline.producer_get_barrier(load_mma_K_producer_state)

                    # Load K tile
                    cute.copy(
                        tma_atom_K,
                        tKgK[None, load_iter_index],
                        tKsK[None, load_mma_K_producer_state.index],
                        tma_bar_ptr=K_tma_barrier,
                    )

                    load_mma_K_producer_state.advance()

                    # Load Q and LSE
                    for h_r_idx in cutlass.range(cute.size(tQgQ, mode=[1])):

                        load_compute_LSE_pipeline.producer_acquire(load_compute_LSE_producer_state)

                        # Load LSE
                        thread_idx = tidx % self.threads_per_warp

                        async_copy_num_elts = sLSE.shape[0] // self.threads_per_warp
                        atom_async_copy = cute.make_copy_atom(
                            cpasync.CopyG2SOp(cache_mode=cute.nvgpu.LoadCacheMode.ALWAYS),
                            self.acc_dtype,
                            num_bits_per_copy=self.acc_dtype.width,
                        )

                        sLSE_for_copy = cute.flat_divide(sLSE, (1,))
                        LSE_for_copy = cute.flat_divide(LSE, (1,))
                        LSE_idx_offset = block_offset[0] * LSE.stride[0]
                        for i in cutlass.range_constexpr(async_copy_num_elts):
                            LSE_idx = self.mma_tiler[0] * bidx + thread_idx * async_copy_num_elts
                            if cute.elem_less(LSE_idx + i, cur_s_q):
                                cute.copy(
                                    atom_async_copy,
                                    LSE_for_copy[
                                        None,
                                        LSE_idx_offset + LSE_idx + i,
                                        h_r_idx,
                                        bidy,
                                        bidz,
                                    ],
                                    sLSE_for_copy[
                                        None,
                                        thread_idx * async_copy_num_elts + i,
                                        load_compute_LSE_producer_state.index,
                                    ],
                                )
                            else:
                                sLSE_for_copy[
                                    None,
                                    thread_idx * async_copy_num_elts + i,
                                    load_compute_LSE_producer_state.index,
                                ].fill(0.0)

                        load_compute_LSE_pipeline.producer_commit(load_compute_LSE_producer_state)
                        load_compute_LSE_producer_state.advance()

                        # Wait for Q to be empty
                        load_mma_Q_pipeline.producer_acquire(load_mma_Q_producer_state)
                        Q_tma_barrier = load_mma_Q_pipeline.producer_get_barrier(load_mma_Q_producer_state)

                        # Load Q tile
                        cute.copy(
                            tma_atom_Q,
                            tQgQ[None, h_r_idx],
                            tQsQ[None, load_mma_Q_producer_state.index],
                            tma_bar_ptr=Q_tma_barrier,
                        )

                        load_mma_Q_producer_state.advance()

                    load_iter_count -= 1
                    load_iter_index += 1

            # MMA WARP
            if warp_idx == self.mma_warp_id:
                # TODO: reconfig regs
                cute.arch.setmaxregister_decrease(self.num_regs_other)

                num_tmem_cols = 512
                cute.arch.alloc_tmem(num_tmem_cols, storage.tmem_holding_buf.ptr)
                cute.arch.barrier(
                    barrier_id=self.tmem_alloc_sync_bar_id,
                    number_of_threads=self.threads_per_warp,
                )

                load_mma_Q_consumer_state = pipeline.make_pipeline_state(pipeline.PipelineUserType.Consumer, self.load_mma_Q_stage)
                load_mma_K_consumer_state = pipeline.make_pipeline_state(pipeline.PipelineUserType.Consumer, self.load_mma_K_stage)
                mma_compute_S_producer_state = pipeline.make_pipeline_state(pipeline.PipelineUserType.Producer, self.mma_compute_S_stage)

                while mma_iter_count > 0:
                    # Wait for K to be full
                    load_mma_K_pipeline.consumer_wait(load_mma_K_consumer_state)

                    for h_r_idx in cutlass.range(cute.size(tQgQ, mode=[1])):
                        # Wait for Q to be full
                        load_mma_Q_pipeline.consumer_wait(load_mma_Q_consumer_state)
                        mma_compute_S_pipeline.producer_acquire(mma_compute_S_producer_state)

                        QK_tiled_mma.set(cute.nvgpu.tcgen05.Field.ACCUMULATE, False)
                        for k_block_idx in cutlass.range_constexpr(cute.size(tSrQ, mode=[2])):
                            cute.gemm(
                                QK_tiled_mma,
                                tStS,
                                tSrQ[
                                    None,
                                    None,
                                    k_block_idx,
                                    load_mma_Q_consumer_state.index,
                                ],
                                tSrK[
                                    None,
                                    None,
                                    k_block_idx,
                                    load_mma_K_consumer_state.index,
                                ],
                                tStS,
                            )
                            QK_tiled_mma.set(cute.nvgpu.tcgen05.Field.ACCUMULATE, True)

                        mma_compute_S_pipeline.producer_commit(mma_compute_S_producer_state)
                        mma_compute_S_producer_state.advance()

                        load_mma_Q_pipeline.consumer_release(load_mma_Q_consumer_state)
                        load_mma_Q_consumer_state.advance()

                    mma_iter_count -= 1
                    mma_iter_index += 1

                    load_mma_K_pipeline.consumer_release(load_mma_K_consumer_state)
                    load_mma_K_consumer_state.advance()

            # COMPUTE WARP
            if warp_idx in self.compute_warp_id:
                cute.arch.setmaxregister_increase(self.num_regs_compute)

                mma_compute_S_consumer_state = pipeline.make_pipeline_state(pipeline.PipelineUserType.Consumer, self.mma_compute_S_stage)
                load_compute_LSE_consumer_state = pipeline.make_pipeline_state(pipeline.PipelineUserType.Consumer, self.load_compute_LSE_stage)

                load_compute_LSE_pipeline.consumer_wait(load_compute_LSE_consumer_state)
                thread_idx = tidx % (self.threads_per_warp * self.num_compute_warps)

                heap_size_ref = cute.make_rmem_tensor((1,), Int32)
                heap_size_ref[0] = 0
                # # Create temporary register heaps for computation
                scores_heap_rf = cute.make_rmem_tensor(((4, self.k_value // 4), 1, 1), Float32)
                idx_heap_rf = cute.make_rmem_tensor(((4, self.k_value // 4), 1, 1), Int32)

                tmem_load_atom = cute.make_copy_atom(
                    tcgen05.Ld32x32bOp(tcgen05.copy.Repetition(32)),
                    self.acc_dtype,
                )
                tmem_store_atom = cute.make_copy_atom(
                    tcgen05.St32x32bOp(tcgen05.copy.Repetition(32)),
                    self.acc_dtype,
                )

                cS = cute.make_identity_tensor((self.mma_tiler[0], self.mma_tiler[1]))

                comp_tile_size = 32
                tStS_tiled = cute.logical_divide(tStS, cute.make_layout((self.mma_tiler[0], comp_tile_size)))
                tStS_compute_tiled = cute.logical_divide(tStS_reduce, cute.make_layout((self.mma_tiler[0], comp_tile_size)))
                cS_tiled = cute.logical_divide(cS, cute.make_layout((self.mma_tiler[0], comp_tile_size)))

                tStS_slice = tStS_tiled[None, 0]  # ((128, 16), 8)
                tStS_compute_slice = tStS_compute_tiled[None, 0]

                # (MMA_M, MMA_N)
                tiled_t2r = tcgen05.make_tmem_copy(tmem_load_atom, tStS_slice)
                thr_t2r = tiled_t2r.get_slice(thread_idx)

                tTR_cS = thr_t2r.partition_D(cS_tiled)
                tTR_tS = thr_t2r.partition_S(tStS_tiled)
                tTR_tS_compute = thr_t2r.partition_S(tStS_compute_tiled)
                tTR_rS = cute.make_rmem_tensor(tTR_cS[None, None, 0].shape, self.acc_dtype)
                tTR_rS_compute = cute.make_rmem_tensor(tTR_cS[None, None, 0].shape, self.acc_dtype)

                tiled_r2t = tcgen05.make_tmem_copy(tmem_store_atom, tStS_compute_slice)
                thr_r2t = tiled_r2t.get_slice(thread_idx)
                tRT_tS_compute = thr_r2t.partition_D(tStS_compute_tiled)

                tiled_t2r_reduce = tcgen05.make_tmem_copy(tmem_load_atom, tStS[(None, None), 0, 0])
                thr_t2r_reduce = tiled_t2r_reduce.get_slice(thread_idx)
                tTR_tS_reduce = thr_t2r_reduce.partition_S(tStS_reduce[(None, None), 0, 0])
                tTR_cS_reduce = thr_t2r_reduce.partition_D(cS)
                tTR_rS_reduce = cute.make_rmem_tensor(tTR_cS_reduce.shape, self.acc_dtype)

                tmp = cute.make_rmem_tensor((self.mma_tiler[1] // self.num_elem_for_reduction), self.acc_dtype)

                while compute_iter_count > 0:

                    for h_r_idx in range(cute.size(tQgQ, mode=[1])):
                        mma_compute_S_pipeline.consumer_wait(mma_compute_S_consumer_state)

                        # TODO: Added this as we should wait for the producer to load
                        load_compute_LSE_pipeline.consumer_wait(load_compute_LSE_consumer_state)

                        for sub_tile in cutlass.range(self.mma_tiler[1] // comp_tile_size):
                            tTR_tS_sub_tile = tTR_tS[None, None, sub_tile]
                            tTR_tS_compute_sub_tile = tTR_tS_compute[None, None, sub_tile]
                            tRT_tS_compute_sub_tile = tRT_tS_compute[None, None, sub_tile]
                            tTR_cS_sub_tile = tTR_cS[None, None, sub_tile]

                            # Copy S from tmem to rmem
                            cute.copy(tiled_t2r, tTR_tS_sub_tile, tTR_rS)

                            is_residual_k = compute_iter_index * self.mma_tiler[1] + self.mma_tiler[1] > cur_s_k

                            leading_causal_masking = cutlass.Boolean(False)
                            if cutlass.const_expr(self.is_causal):
                                leading_causal_masking = (
                                    (compute_iter_index + 1) * self.mma_tiler[1] + 1
                                ) * self.compress_block_sliding_stride - 1 > bidx * self.mma_tiler[0]
                                leading_causal_masking = cute.arch.shuffle_sync(leading_causal_masking, 0)
                            trailing_residual_masking = cutlass.Boolean(False)
                            trailing_residual_masking = is_residual_k
                            trailing_residual_masking = cute.arch.shuffle_sync(trailing_residual_masking, 0)

                            is_masked_tile = leading_causal_masking or trailing_residual_masking

                            # Apply mask
                            if is_masked_tile:
                                for i in cutlass.range(cute.size(tTR_rS), unroll_full=True):
                                    q_idx = cute.get(tTR_cS_sub_tile[i], mode=[0]) + bidx * self.mma_tiler[0]
                                    k_block_idx = cute.get(tTR_cS_sub_tile[i], mode=[1]) + compute_iter_index * self.mma_tiler[1]

                                    if is_masked_tile:
                                        if cutlass.const_expr(self.is_causal):
                                            k_idx = (k_block_idx + 1) * self.compress_block_sliding_stride - 1
                                            if k_idx > q_idx:
                                                tTR_rS[i] = -cutlass.Float32.inf
                                            if q_idx > cur_s_q or k_block_idx > cur_s_k:
                                                tTR_rS[i] = -cutlass.Float32.inf
                                        else:
                                            if q_idx > cur_s_q or k_block_idx > cur_s_k:
                                                tTR_rS[i] = -cutlass.Float32.inf

                            # P = exp2(S * softmax_scale_log2_e - LSE),
                            # LSE should be set negative before and has be already multiplied by log2_e

                            # Copy S_reduce from tmem to rmem
                            cute.copy(tiled_t2r, tTR_tS_compute_sub_tile, tTR_rS_compute)

                            for i in cutlass.range(0, cute.size(tTR_rS, mode=[0]), 2):
                                lse = (
                                    sLSE[
                                        cute.get(tTR_cS_sub_tile[i], mode=[0]),
                                        load_compute_LSE_consumer_state.index,
                                    ],
                                    sLSE[
                                        cute.get(tTR_cS_sub_tile[i + 1], mode=[0]),
                                        load_compute_LSE_consumer_state.index,
                                    ],
                                )

                                tTR_rS[i], tTR_rS[i + 1] = cute.arch.fma_packed_f32x2(
                                    (tTR_rS[i], tTR_rS[i + 1]),
                                    (
                                        softmax_scale_log2_e,
                                        softmax_scale_log2_e,
                                    ),
                                    lse,
                                )
                                tTR_rS[i] = cute.math.exp2(tTR_rS[i], fastmath=True)
                                tTR_rS[i + 1] = cute.math.exp2(tTR_rS[i + 1], fastmath=True)

                                if h_r_idx == 0:
                                    tTR_rS_compute[i], tTR_rS_compute[i + 1] = cute.arch.add_packed_f32x2(
                                        (0.0, 0.0),
                                        (tTR_rS[i], tTR_rS[i + 1]),
                                    )
                                else:
                                    tTR_rS_compute[i], tTR_rS_compute[i + 1] = cute.arch.add_packed_f32x2(
                                        (tTR_rS_compute[i], tTR_rS_compute[i + 1]),
                                        (tTR_rS[i], tTR_rS[i + 1]),
                                    )

                            cute.arch.fence_view_async_tmem_load()
                            cute.arch.barrier(
                                barrier_id=self.compute_sync_bar_id,
                                number_of_threads=self.num_compute_warps * self.threads_per_warp,
                            )

                            # Copy tS_reduce back to tmem
                            cute.copy(tiled_r2t, tTR_rS_compute, tRT_tS_compute_sub_tile)

                            cute.arch.fence_view_async_tmem_store()

                        load_compute_LSE_pipeline.consumer_release(load_compute_LSE_consumer_state)
                        load_compute_LSE_consumer_state.advance()

                        mma_compute_S_pipeline.consumer_release(mma_compute_S_consumer_state)
                        mma_compute_S_consumer_state.advance()

                    # Reduce
                    cute.copy(tiled_t2r_reduce, tTR_tS_reduce, tTR_rS_reduce)

                    # Contraint: self.mma_tiler[1] % self.num_elem_for_reduction == 0
                    tTR_rS_reduce_reshape = cute.composition(
                        tTR_rS_reduce,
                        cute.make_layout(
                            (
                                self.mma_tiler[1] // self.num_elem_for_reduction,
                                self.num_elem_for_reduction,
                            ),
                            stride=(self.num_elem_for_reduction, 1),
                        ),
                    )
                    tTR_rS_reduce_vec = tTR_rS_reduce_reshape.load()
                    tTR_rS_reduce_sum_vec = tTR_rS_reduce_vec.reduce(
                        cute.ReductionOp.ADD,
                        self.acc_dtype(0.0),
                        reduction_profile=(None, 1),
                    )

                    tmp.store(tTR_rS_reduce_sum_vec)

                    q_row_idx = bidx * self.mma_tiler[0] + thread_idx
                    # TODO: make sure the self.mma_tile[1] // self.num_elem_for_reduction is always larger than self.k_value
                    if compute_iter_index == k_tile_idx:
                        self.topk_step(
                            tmp,
                            scores_heap_rf,
                            idx_heap_rf,
                            q_row_idx,
                            compute_iter_index,
                            self.k_value,
                        )
                    else:
                        self.topk_step(
                            tmp,
                            scores_heap_rf,
                            idx_heap_rf,
                            q_row_idx,
                            compute_iter_index,
                            0,
                        )

                    compute_iter_count -= 1
                    compute_iter_index += 1

                # (s_q, k_value, (1, h_k, b))
                gTopk_scores = cute.flat_divide(mTopk_scores, (self.epi_tile[0], self.k_value))
                gTopk_indices = cute.flat_divide(mTopk_indices, (self.epi_tile[0], self.k_value))
                gTopk_scores = gTopk_scores[None, None, bidx, 0, (0, bidy, bidz)]
                gTopk_indices = gTopk_indices[None, None, bidx, 0, (0, bidy, bidz)]
                cTopk = cute.make_identity_tensor((self.epi_tile[0], self.k_value))
                cTopk = cute.domain_offset((bidx * self.epi_tile[0], 0), cTopk)

                copy_atom = cute.make_copy_atom(
                    cute.nvgpu.CopyUniversalOp(),
                    self.acc_dtype,
                    num_bits_per_copy=128,
                )

                thr_layout = cute.make_layout(
                    (self.num_compute_warps * self.threads_per_warp, 1),
                    stride=(1, 1),
                )
                val_layout = cute.make_layout(
                    (1, self.k_value),
                    stride=(self.k_value, 1),
                )
                copy_op = cute.make_tiled_copy_tv(
                    copy_atom,
                    thr_layout,
                    val_layout,
                )

                thr_copy = copy_op.get_slice(thread_idx)

                tgTopk_scores = thr_copy.partition_D(gTopk_scores)
                tgTopk_indices = thr_copy.partition_D(gTopk_indices)
                tcTopk = thr_copy.partition_D(cTopk)

                pred_shape = (tcTopk.shape[0][1], tcTopk.shape[1], tcTopk.shape[2])
                preds = cute.make_rmem_tensor(pred_shape, cutlass.Boolean)
                for v in cutlass.range_constexpr(preds.shape[0]):
                    for m in cutlass.range_constexpr(preds.shape[1]):
                        for n in cutlass.range_constexpr(preds.shape[2]):
                            lhs = tcTopk[v, m, n]
                            val = cute.elem_less(lhs, (cur_s_q, self.k_value))
                            preds[v, m, n] = val

                cute.copy(copy_atom, scores_heap_rf, tgTopk_scores, pred=preds)
                cute.copy(copy_atom, idx_heap_rf, tgTopk_indices, pred=preds)

                cute.arch.barrier(
                    barrier_id=self.compute_sync_bar_id,
                    number_of_threads=self.num_compute_warps * self.threads_per_warp,
                )

                if warp_idx % self.num_compute_warps == 0:
                    tmem_ptr = cute.arch.retrieve_tmem_ptr(
                        self.acc_dtype,
                        alignment=16,
                        ptr_to_buffer_holding_addr=storage.tmem_holding_buf.ptr,
                    )
                    cute.arch.dealloc_tmem(tmem_ptr, 512)

    @cute.jit
    def topk_step(
        self,
        tiled_scores: cute.Tensor,  # (reduced_tile) probabilities
        scores_heap_rf: cute.Tensor,
        idx_heap_rf: cute.Tensor,
        query_index: Int32,
        k_tile_idx: Int32,
        heap_size: int,
    ):

        q_selection_block_idx = (query_index + 1) // self.selection_block_size
        blocks_per_tile = self.mma_tiler[1] // self.num_elem_for_reduction

        score = 0.0
        gmem_idx = 0
        is_valid = cutlass.Boolean(False)

        for i in cutlass.range_constexpr(heap_size):
            score = tiled_scores[i]
            gmem_idx = i + k_tile_idx * blocks_per_tile

            is_valid = (gmem_idx > 0) and (gmem_idx < q_selection_block_idx - 2)

            val_score = score if is_valid else -cutlass.Float32.inf
            val_index = gmem_idx if is_valid else -1
            scores_heap_rf[i] = val_score
            idx_heap_rf[i] = val_index

        for i in cutlass.range(heap_size, cute.size(tiled_scores)):
            score = tiled_scores[i]
            gmem_idx = i + k_tile_idx * blocks_per_tile

            # Decide if candidate is valid for consideration
            is_valid = (gmem_idx > 0) and (gmem_idx < q_selection_block_idx - 2)

            # Full: branchless min-scan to find current minimum (with tie-break)
            min_score = scores_heap_rf[0]
            min_index = idx_heap_rf[0]
            for j in cutlass.range(1, self.k_value, unroll_full=True):
                s = scores_heap_rf[j]
                t = idx_heap_rf[j]
                is_smaller_score = s < min_score
                is_tie_break = (s == min_score) and (t > min_index)
                if is_smaller_score or is_tie_break:
                    # swap to make sure the min_slot is always 0
                    scores_heap_rf[0], scores_heap_rf[j] = s, min_score
                    idx_heap_rf[0], idx_heap_rf[j] = t, min_index
                    min_score = s
                    min_index = t

            # Decide replacement under tie-break rules
            if is_valid:
                is_larger_score = score > min_score
                is_tie_better = (score == min_score) and (gmem_idx < min_index)
                if is_larger_score or is_tie_better:
                    scores_heap_rf[0] = score
                    idx_heap_rf[0] = gmem_idx



# Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

from .grouped_gemm_swiglu.api import (
    GroupedGemmSwigluSm100,
    grouped_gemm_swiglu_wrapper_sm100,
)

from .grouped_gemm_dswiglu.api import (
    GroupedGemmDswigluSm100,
    grouped_gemm_dswiglu_wrapper_sm100,
)

from .grouped_gemm_quant.api import (
    GroupedGemmQuantSm100,
    grouped_gemm_quant_wrapper_sm100,
)

from .grouped_gemm_srelu.api import (
    GroupedGemmSreluSm100,
    grouped_gemm_srelu_wrapper_sm100,
)

from .grouped_gemm_dsrelu.api import (
    GroupedGemmDsreluSm100,
    grouped_gemm_dsrelu_wrapper_sm100,
)

from .grouped_gemm_glu.api import (
    GroupedGemmGluSm100,
    grouped_gemm_glu_wrapper_sm100,
)

from .grouped_gemm_glu_hadamard.api import (
    GroupedGemmGluHadamardSm100,
    grouped_gemm_glu_hadamard_wrapper_sm100,
)

from .grouped_gemm_dglu.api import (
    GroupedGemmDgluSm100,
    grouped_gemm_dglu_wrapper_sm100,
)

from .grouped_gemm_wgrad.api import (
    GroupedGemmWgradSm100,
    grouped_gemm_wgrad_wrapper_sm100,
)

__all__ = [
    "GroupedGemmSwigluSm100",
    "grouped_gemm_swiglu_wrapper_sm100",
    "GroupedGemmDswigluSm100",
    "grouped_gemm_dswiglu_wrapper_sm100",
    "GroupedGemmQuantSm100",
    "grouped_gemm_quant_wrapper_sm100",
    "GroupedGemmSreluSm100",
    "grouped_gemm_srelu_wrapper_sm100",
    "GroupedGemmDsreluSm100",
    "grouped_gemm_dsrelu_wrapper_sm100",
    "GroupedGemmGluSm100",
    "grouped_gemm_glu_wrapper_sm100",
    "GroupedGemmGluHadamardSm100",
    "grouped_gemm_glu_hadamard_wrapper_sm100",
    "GroupedGemmDgluSm100",
    "grouped_gemm_dglu_wrapper_sm100",
    "GroupedGemmWgradSm100",
    "grouped_gemm_wgrad_wrapper_sm100",
]



# Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""FE API for fused RMSNorm + RHT + per-CTA amax."""

import logging
from typing import Optional

from cuda.bindings import driver as cuda
import cutlass
import cutlass.cute as cute
import torch
from cutlass import Float32
from cutlass.cute.runtime import make_fake_stream

from cudnn.api_base import APIBase, TupleDict

from .kernel import RMSNormRHTAmaxKernel

DEFAULT_NUM_THREADS_BY_N = {
    2048: 128,
    4096: 256,
    7168: 128,
    8192: 512,
    16384: 1024,
    32768: 512,
}
RPC_CANDIDATES = (2, 4, 8)
TARGET_MIN_CTAS = 148


def best_num_threads(n: int) -> Optional[int]:
    for num_threads in (1024, 512, 256, 128, 64):
        if n % num_threads != 0:
            continue
        ept = n // num_threads
        if ept >= 8 and ept % 8 == 0:
            return num_threads
    return None


def pick_rows_per_cta(m: int) -> int:
    for rows_per_cta in reversed(RPC_CANDIDATES):
        if m % rows_per_cta != 0:
            continue
        num_ctas = m // rows_per_cta
        if num_ctas >= TARGET_MIN_CTAS:
            return rows_per_cta
    return RPC_CANDIDATES[0]


class RmsNormRhtAmaxSm100(APIBase):
    """Class API for the RMSNorm + RHT + amax kernel."""

    def __init__(
        self,
        sample_x: torch.Tensor,
        sample_w: torch.Tensor,
        sample_o: torch.Tensor,
        sample_amax: torch.Tensor,
        eps: float = 1e-5,
        num_threads: Optional[int] = None,
        rows_per_cta: Optional[int] = None,
    ):
        super().__init__()

        self._warn_experimental_api()

        self.x_desc = self._unpad_tensor_to_ndim(self._make_tensor_desc(sample_x, name="sample_x"), 2, "sample_x")
        self.w_desc = self._unpad_tensor_to_ndim(self._make_tensor_desc(sample_w, name="sample_w"), 1, "sample_w")
        self.o_desc = self._unpad_tensor_to_ndim(self._make_tensor_desc(sample_o, name="sample_o"), 2, "sample_o")
        self.amax_desc = self._unpad_tensor_to_ndim(self._make_tensor_desc(sample_amax, name="sample_amax"), 1, "sample_amax")

        self.eps = eps
        self.requested_num_threads = num_threads
        self.requested_rows_per_cta = rows_per_cta
        self.num_threads = None
        self.rows_per_cta = None
        self.n = None

    def check_support(self) -> bool:
        m, n = self._tensor_shape(self.x_desc, name="sample_x")
        w_n = self._tensor_shape(self.w_desc, name="sample_w")[0]
        o_m, o_n = self._tensor_shape(self.o_desc, name="sample_o")

        self._check_tensor_shape(self.x_desc, (m, n), "X")
        self._check_tensor_shape(self.w_desc, (n,), "W")
        self._check_tensor_shape(self.o_desc, (m, n), "O")
        self._value_error_if(w_n != n, f"W length must match X hidden dimension, got {w_n} and {n}")
        self._value_error_if((n % 16) != 0, f"N must be divisible by 16 for the Hadamard block size, got {n}")
        self._value_error_if(o_m != m or o_n != n, f"O shape must match X shape, got {(o_m, o_n)} and {(m, n)}")

        self._check_tensor_stride(self.x_desc, stride=(n, 1), name="X", extra_error_msg="X must be row-major contiguous")
        self._check_tensor_stride(self.w_desc, stride=(1,), name="W", extra_error_msg="W must be contiguous")
        self._check_tensor_stride(self.o_desc, stride=(n, 1), name="O", extra_error_msg="O must be row-major contiguous")

        self._check_dtype(self.x_desc, dtype=torch.bfloat16, name="X")
        self._check_dtype(self.w_desc, dtype=torch.bfloat16, name="W")
        self._check_dtype(self.o_desc, dtype=torch.bfloat16, name="O")
        self._check_dtype(self.amax_desc, dtype=torch.float32, name="Amax")

        resolved_num_threads = self.requested_num_threads
        if resolved_num_threads is None:
            resolved_num_threads = DEFAULT_NUM_THREADS_BY_N.get(n, best_num_threads(n))
        self._value_error_if(resolved_num_threads is None, f"No valid num_threads found for N={n}")
        self._value_error_if(resolved_num_threads <= 0, f"num_threads must be positive, got {resolved_num_threads}")
        self._value_error_if(
            (resolved_num_threads % 32) != 0,
            f"num_threads must be warp-aligned, got {resolved_num_threads}",
        )
        self._value_error_if(
            resolved_num_threads > 1024,
            f"num_threads must not exceed the CUDA block size limit, got {resolved_num_threads}",
        )

        resolved_rows_per_cta = self.requested_rows_per_cta
        if resolved_rows_per_cta is None:
            resolved_rows_per_cta = pick_rows_per_cta(m)

        self._value_error_if(m % resolved_rows_per_cta != 0, f"M must be divisible by rows_per_cta, got M={m}, rows_per_cta={resolved_rows_per_cta}")
        self._value_error_if(n % resolved_num_threads != 0, f"N={n} must be divisible by num_threads={resolved_num_threads}")

        ept = n // resolved_num_threads
        self._value_error_if(ept < 8 or ept % 8 != 0, f"EPT={ept} must be >= 8 and divisible by 8")

        expected_num_ctas = m // resolved_rows_per_cta
        self._check_tensor_shape(self.amax_desc, (expected_num_ctas,), "Amax")

        self._runtime_error_if(not torch.cuda.is_available(), "CUDA is not available")
        major, minor = torch.cuda.get_device_capability(self.x_desc.device)
        compute_capability = major * 10 + minor
        self._runtime_error_if(
            compute_capability < 100,
            f"RmsNormRhtAmaxSm100 requires SM100+, found SM{compute_capability}",
        )

        self.num_threads = resolved_num_threads
        self.rows_per_cta = resolved_rows_per_cta
        self.n = n
        self._is_supported = True
        return True

    def compile(self) -> None:
        self._ensure_support_checked()
        if self._compiled_kernel is not None:
            return

        kernel = RMSNormRHTAmaxKernel(
            n=self.n,
            num_threads=self.num_threads,
            eps=self.eps,
            rows_per_cta=self.rows_per_cta,
        )

        valid_m = cute.sym_int(divisibility=self.rows_per_cta)

        fake_x_tensor = self._make_fake_cute_compact_tensor(
            dtype=self.x_desc.dtype,
            shape=(valid_m, self.n),
            stride_order=self.x_desc.stride_order,
            dynamic_mode=None,
            divisibility=self.rows_per_cta,
        )
        fake_w_tensor = self._make_fake_cute_tensor_from_desc(self.w_desc, assumed_align=16)
        fake_o_tensor = self._make_fake_cute_compact_tensor(
            dtype=self.o_desc.dtype,
            shape=(valid_m, self.n),
            stride_order=self.o_desc.stride_order,
            dynamic_mode=None,
            divisibility=self.rows_per_cta,
        )
        fake_num_ctas = cute.sym_int()
        fake_amax_tensor = self._make_fake_cute_tensor(
            dtype=self.amax_desc.dtype,
            shape=(fake_num_ctas,),
            stride=self.amax_desc.stride,
            assumed_align=16,
        )
        fake_stream = make_fake_stream(use_tvm_ffi_env_stream=False)

        compiled_kernel = cute.compile(
            kernel,
            fake_x_tensor,
            fake_w_tensor,
            fake_o_tensor,
            fake_amax_tensor,
            Float32(self.eps),
            fake_stream,
            options="--enable-tvm-ffi",
        )

        def tensor_api(
            x_tensor: torch.Tensor,
            w_tensor: torch.Tensor,
            o_tensor: torch.Tensor,
            amax_tensor: torch.Tensor,
            stream: cuda.CUstream,
        ) -> None:
            compiled_kernel(
                x_tensor,
                w_tensor,
                o_tensor,
                amax_tensor,
                Float32(self.eps),
                stream,
            )

        self._compiled_kernel = tensor_api

    def execute(
        self,
        x_tensor: torch.Tensor,
        w_tensor: torch.Tensor,
        o_tensor: torch.Tensor,
        amax_tensor: torch.Tensor,
        current_stream: Optional[cuda.CUstream] = None,
    ) -> None:
        self._runtime_error_if(self._compiled_kernel is None, "RmsNormRhtAmaxSm100 kernel not compiled; call compile() first")

        x_tensor = self._unpad_tensor_to_ndim(x_tensor, 2, "x_tensor")
        w_tensor = self._unpad_tensor_to_ndim(w_tensor, 1, "w_tensor")
        o_tensor = self._unpad_tensor_to_ndim(o_tensor, 2, "o_tensor")
        amax_tensor = self._unpad_tensor_to_ndim(amax_tensor, 1, "amax_tensor")

        if current_stream is None:
            current_stream = cuda.CUstream(torch.cuda.current_stream(x_tensor.device).cuda_stream)

        self._compiled_kernel(
            x_tensor=x_tensor,
            w_tensor=w_tensor,
            o_tensor=o_tensor,
            amax_tensor=amax_tensor,
            stream=current_stream,
        )


_logger = logging.getLogger(__name__)
_cache_of_RmsNormRhtAmaxSm100Objects = {}


def rmsnorm_rht_amax_wrapper_sm100(
    x_tensor: torch.Tensor,
    w_tensor: torch.Tensor,
    eps: float = 1e-5,
    num_threads: Optional[int] = None,
    rows_per_cta: Optional[int] = None,
    current_stream: Optional[cuda.CUstream] = None,
) -> TupleDict:
    """High-level wrapper for the RMSNorm + RHT + per-CTA amax kernel."""

    x_tensor = x_tensor.squeeze(-1) if x_tensor.ndim == 3 and x_tensor.shape[-1] == 1 else x_tensor
    w_tensor = w_tensor.squeeze(-1) if w_tensor.ndim == 2 and w_tensor.shape[-1] == 1 else w_tensor

    m, n = x_tensor.shape
    resolved_num_threads = num_threads if num_threads is not None else DEFAULT_NUM_THREADS_BY_N.get(n, best_num_threads(n))
    if resolved_num_threads is None:
        raise ValueError(f"No valid num_threads found for N={n}")
    resolved_rows_per_cta = rows_per_cta if rows_per_cta is not None else pick_rows_per_cta(m)
    if m % resolved_rows_per_cta != 0:
        raise ValueError(f"M must be divisible by rows_per_cta, got M={m}, rows_per_cta={resolved_rows_per_cta}")

    o_tensor = torch.empty_like(x_tensor)
    amax_tensor = torch.full((m // resolved_rows_per_cta,), float("-inf"), dtype=torch.float32, device=x_tensor.device)

    cache_key = (
        n,
        x_tensor.dtype,
        w_tensor.dtype,
        o_tensor.dtype,
        tuple(x_tensor.stride()),
        tuple(w_tensor.stride()),
        tuple(o_tensor.stride()),
        eps,
        resolved_num_threads,
        resolved_rows_per_cta,
    )

    if cache_key in _cache_of_RmsNormRhtAmaxSm100Objects:
        api = _cache_of_RmsNormRhtAmaxSm100Objects[cache_key]
    else:
        api = RmsNormRhtAmaxSm100(
            sample_x=x_tensor,
            sample_w=w_tensor,
            sample_o=o_tensor,
            sample_amax=amax_tensor,
            eps=eps,
            num_threads=resolved_num_threads,
            rows_per_cta=resolved_rows_per_cta,
        )
        assert api.check_support(), "Unsupported configuration"
        api.compile()
        _cache_of_RmsNormRhtAmaxSm100Objects[cache_key] = api

    api.execute(
        x_tensor=x_tensor,
        w_tensor=w_tensor,
        o_tensor=o_tensor,
        amax_tensor=amax_tensor,
        current_stream=current_stream,
    )

    return TupleDict(o_tensor=o_tensor, amax_tensor=amax_tensor)



# Copyright (c) 2025, Tri Dao.
# Copyright (c) 2026, Jerry Chen
"""SM90 CuTe DSL primitives shared by DSA kernels."""

import math
from typing import Type, Callable, overload

import cutlass
import cutlass.cute as cute

from cutlass import Float32, const_expr
from cutlass.cutlass_dsl import T, dsl_user_op
from cutlass._mlir.dialects import nvvm, llvm


def get_smem_store_atom(arch: cutlass.Constexpr[int], element_type: Type[cute.Numeric], transpose: bool = False) -> cute.CopyAtom:
    if const_expr(arch < 90 or element_type.width != 16):
        return cute.make_copy_atom(
            cute.nvgpu.CopyUniversalOp(),
            element_type,
            num_bits_per_copy=2 * element_type.width,
        )
    else:
        return cute.make_copy_atom(
            cute.nvgpu.warp.StMatrix8x8x16bOp(transpose=transpose, num_matrices=4),
            element_type,
        )


@cute.jit
def warp_reduce(
    val: cute.TensorSSA | cute.Numeric,
    op: Callable,
    width: cutlass.Constexpr[int] = cute.arch.WARP_SIZE,
) -> cute.TensorSSA | cute.Numeric:
    if const_expr(isinstance(val, cute.TensorSSA)):
        res = cute.make_rmem_tensor(val.shape, val.dtype)
        res.store(val)
        for i in cutlass.range_constexpr(cute.size(val.shape)):
            res[i] = warp_reduce(res[i], op, width)
        return res.load()
    else:
        for i in cutlass.range_constexpr(int(math.log2(width))):
            val = op(val, cute.arch.shuffle_sync_bfly(val, offset=1 << i))
    return val


def convert_layout_acc_mn(acc_layout: cute.Layout, transpose: bool = False) -> cute.Layout:
    """
    For Sm80, convert ((2, 2), MMA_M, MMA_N, ...) to ((2, MMA_M), (2, MMA_N), ...).
    For Sm90, convert ((2, 2, V), MMA_M, MMA_N, ...) to ((2, MMA_M), (2, V, MMA_N), ...).
    """
    acc_layout_col_major = cute.make_layout(acc_layout.shape)
    shape = (
        (acc_layout_col_major.shape[0][1], acc_layout_col_major.shape[1]),  # MMA_M
        (
            acc_layout_col_major.shape[0][0],
            *acc_layout_col_major.shape[0][2:],
            acc_layout_col_major.shape[2],
        ),  # MMA_N
        *acc_layout_col_major.shape[3:],
    )
    stride = (
        (acc_layout_col_major.stride[0][1], acc_layout_col_major.stride[1]),  # MMA_M
        (
            acc_layout_col_major.stride[0][0],
            *acc_layout_col_major.stride[0][2:],
            acc_layout_col_major.stride[2],
        ),  # MMA_N
        *acc_layout_col_major.stride[3:],
    )
    if const_expr(transpose):
        shape = (shape[1], shape[0], *shape[2:])
        stride = (stride[1], stride[0], *stride[2:])
    acc_layout_mn = cute.make_layout(shape, stride=stride)
    return cute.composition(acc_layout, acc_layout_mn)


def make_acc_tensor_mn_view(acc: cute.Tensor, transpose: bool = False) -> cute.Tensor:
    return cute.make_tensor(acc.iterator, convert_layout_acc_mn(acc.layout, transpose=transpose))


@cute.jit
def convert_layout_acc_frgA(acc_layout: cute.Layout) -> cute.Layout:
    # For back to back gemm, convert layout of acc0 to gemm 1 accept layout.
    # For Sm80, as the mma instruction shape is 16x8x16, we need to convert from (4, MMA_M, MMA_N) to ((4, 2), MMA_M, MMA_N / 2)
    # For Sm90, FP16/BF16, convert acc_layout from ((2, 2, N / 8), MMA_M, MMA_N) to ((2, 2, 2), MMA_M, (N / 16, MMA_N))
    # TODO: Sm90 FP8
    if const_expr(cute.rank(acc_layout.shape[0]) == 3):  # Sm90
        l = cute.logical_divide(acc_layout, ((None, None, 2), None, None))  # ((2, 2, (2, N / 16)), MMA_M, MMA_N)
        rA_mma_view = cute.make_layout(
            (
                (l.shape[0][0], l.shape[0][1], l.shape[0][2][0]),
                l.shape[1],
                (l.shape[0][2][1], l.shape[2]),
            ),
            stride=(
                (l.stride[0][0], l.stride[0][1], l.stride[0][2][0]),
                l.stride[1],
                (l.stride[0][2][1], l.stride[2]),
            ),
        )
    else:  # Sm80
        # (4, MMA_M, MMA_N) -> (4, MMA_M, (2, MMA_N / 2))
        l = cute.logical_divide(acc_layout, (None, None, 2))
        rA_mma_view = cute.make_layout(
            (
                (l.shape[0], l.shape[2][0]),
                l.shape[1],
                l.shape[2][1],
            ),
            stride=(
                (l.stride[0], l.stride[2][0]),
                l.stride[1],
                l.stride[2][1],
            ),
        )
    return rA_mma_view


def make_acc_tensor_frgA_view(acc: cute.Tensor) -> cute.Tensor:
    return cute.make_tensor(acc.iterator, convert_layout_acc_frgA(acc.layout))


def select(a: cute.Tensor, mode: list[int]) -> cute.Tensor:
    return cute.make_tensor(a.iterator, cute.select(a.layout, mode))


def transpose_view(a: cute.Tensor) -> cute.Tensor:
    """Transpose the first two dimensions of a tensor on smem."""
    shape = (a.shape[1], a.shape[0], *a.shape[2:])
    order = (1, 0, *range(2, cute.rank(a)))
    return cute.composition(a, cute.make_ordered_layout(shape, order=order))


@dsl_user_op
def log2f(a: float | Float32, *, loc=None, ip=None) -> Float32:
    return Float32(
        llvm.inline_asm(
            T.f32(),
            [Float32(a).ir_value(loc=loc, ip=ip)],
            "lg2.approx.ftz.f32 $0, $1;",
            "=f,f",
            has_side_effects=False,
            is_align_stack=False,
            asm_dialect=llvm.AsmDialect.AD_ATT,
        )
    )


@dsl_user_op
def logf(a: float | Float32, *, loc=None, ip=None) -> Float32:
    return log2f(a, loc=loc, ip=ip) * math.log(2.0)


@dsl_user_op
def atomic_add_fp32(a: float | Float32, gmem_ptr: cute.Pointer, *, loc=None, ip=None) -> None:
    """Scalar f32 atomicAdd: atomically adds `a` to the f32 value at `gmem_ptr`."""
    nvvm.atomicrmw(op=nvvm.AtomicOpKind.FADD, ptr=gmem_ptr.llvm_ptr, a=Float32(a).ir_value())


@dsl_user_op
def atomic_add_fp32x4(
    a0: Float32,
    a1: Float32,
    a2: Float32,
    a3: Float32,
    gmem_ptr: cute.Pointer,
    *,
    loc=None,
    ip=None,
) -> None:
    """Vectorized float4 atomicAdd via atom.relaxed.gpu.global.add.v4.f32.

    Requires `gmem_ptr` to be 16-byte aligned (guaranteed by the 128-bit
    copy atom in r2s_tiled_copy_dKVatomic).
    """
    llvm.inline_asm(
        llvm.StructType.get_literal([T.f32(), T.f32(), T.f32(), T.f32()]),
        [
            gmem_ptr.toint().ir_value(loc=loc, ip=ip),
            Float32(a0).ir_value(loc=loc, ip=ip),
            Float32(a1).ir_value(loc=loc, ip=ip),
            Float32(a2).ir_value(loc=loc, ip=ip),
            Float32(a3).ir_value(loc=loc, ip=ip),
        ],
        "atom.relaxed.gpu.global.add.v4.f32 {$0,$1,$2,$3}, [$4], {$5,$6,$7,$8};",  # relaxed mode
        "=f,=f,=f,=f,l,f,f,f,f",
        has_side_effects=True,
        is_align_stack=False,
        asm_dialect=llvm.AsmDialect.AD_ATT,
        loc=loc,
        ip=ip,
    )


@dsl_user_op
def elem_pointer(x: cute.Tensor, coord: cute.Coord, *, loc=None, ip=None) -> cute.Pointer:
    """Return the gmem pointer to element at `coord` in tensor `x`."""
    return x.iterator + cute.crd2idx(coord, x.layout, loc=loc, ip=ip)


@dsl_user_op
def elem_pointer_packed_i64(
    base_ptr_i64: cute.typing.Int,
    h_idx: cute.typing.Int,
    m_idx: cute.typing.Int,
    seqlen_q: cute.typing.Int,
    elem_type: type,
    memspace: cute.AddressSpace,
    *,
    loc=None,
    ip=None,
) -> cute.Pointer:
    """For packed layout ((qhpkv, seqlen_q),): linear_idx = h_idx * seqlen_q + m_idx."""
    linear_idx = cutlass.Int64(h_idx) * cutlass.Int64(seqlen_q) + cutlass.Int64(m_idx)
    byte_offset = linear_idx * elem_type.width // 8
    return cute.make_ptr(
        elem_type,
        cutlass.Int64(base_ptr_i64) + byte_offset,
        memspace,
        assumed_align=16,
    )


@cute.jit
def predicate_k(tAcA: cute.Tensor, limit: cutlass.Int32) -> cute.Tensor:
    # Only compute predicates for the "k" dimension. For the mn dimension, we will use "if"
    tApA = cute.make_rmem_tensor(
        cute.make_layout(
            (cute.size(tAcA, mode=[0, 1]), cute.size(tAcA, mode=[1]), cute.size(tAcA, mode=[2])),
            stride=(cute.size(tAcA, mode=[2]), 0, 1),
        ),
        cutlass.Boolean,
    )
    for rest_v in cutlass.range_constexpr(tApA.shape[0]):
        for rest_k in cutlass.range_constexpr(tApA.shape[2]):
            tApA[rest_v, 0, rest_k] = cute.elem_less(tAcA[(0, rest_v), 0, rest_k][1], limit)
    return tApA


@cute.jit
def shuffle_sync(
    value: cute.Numeric,
    offset: cute.typing.Int,
    width: cutlass.Constexpr[int] = cute.arch.WARP_SIZE,
) -> cute.Numeric:
    assert value.width % 32 == 0, "value type must be a multiple of 32 bits"
    mask = cute.arch.WARP_SIZE - width
    clamp = cute.arch.WARP_SIZE - 1
    mask_and_clamp = mask << 8 | clamp
    val = cute.make_rmem_tensor(cute.make_layout((1,), stride=(1,)), type(value))
    val[0] = value
    val_i32 = cute.recast_tensor(val, cutlass.Int32)
    for i in cutlass.range_constexpr(cute.size(val_i32)):
        val_i32[i] = cute.arch.shuffle_sync(val_i32[i], offset, mask_and_clamp=mask_and_clamp)
    return val[0]


@dsl_user_op
def cvt_f16x2_f32(a: float | Float32, b: float | Float32, to_dtype: Type, *, loc=None, ip=None) -> cutlass.Int32:
    assert to_dtype in [cutlass.BFloat16, cutlass.Float16], "to_dtype must be BFloat16 or Float16"
    return cutlass.Int32(
        llvm.inline_asm(
            T.i32(),
            [Float32(a).ir_value(loc=loc, ip=ip), Float32(b).ir_value(loc=loc, ip=ip)],
            f"cvt.rn.{'bf16x2' if to_dtype is cutlass.BFloat16 else 'f16x2'}.f32 $0, $2, $1;",
            "=r,f,f",
            has_side_effects=False,
            is_align_stack=False,
            asm_dialect=llvm.AsmDialect.AD_ATT,
        )
    )


@overload
def cvt_f16(src: cute.Tensor, dst: cute.Tensor) -> None: ...


@overload
def cvt_f16(src: cute.Tensor, dtype: Type[cute.Numeric]) -> cute.Tensor: ...


@cute.jit
def cvt_f16(src: cute.Tensor, dst_or_dtype):
    """Convert Float32 tensor to Float16/BFloat16."""
    if const_expr(isinstance(dst_or_dtype, type)):
        dtype = dst_or_dtype
        dst = cute.make_rmem_tensor(src.shape, dtype)
        cvt_f16(src, dst)
        return dst
    else:
        dst = dst_or_dtype
        assert cute.size(dst.shape) == cute.size(src.shape), "dst and src must have the same size"
        assert cute.size(src.shape) % 2 == 0, "src must have an even number of elements"
        assert dst.element_type in [cutlass.BFloat16, cutlass.Float16], "dst must be BFloat16 or Float16"
        assert src.element_type is Float32, "src must be Float32"
        dst_i32 = cute.recast_tensor(dst, cutlass.Int32)
        assert cute.size(dst_i32.shape) * 2 == cute.size(src.shape)
        for i in cutlass.range_constexpr(cute.size(dst_i32)):
            dst_i32[i] = cvt_f16x2_f32(src[2 * i], src[2 * i + 1], dst.element_type)



from .dense_blockscaled_gemm_persistent_amax import (
    Sm100BlockScaledPersistentDenseGemmKernel,
)

from cuda.bindings import driver as cuda
import os
import torch
from typing import Tuple, Optional

import cutlass
import cutlass.cute as cute
from cutlass.cute.runtime import make_fake_stream

from cudnn.datatypes import _convert_to_cutlass_data_type
from cudnn.api_base import APIBase, TensorDesc, TupleDict, is_power_of_2, ceil_div


class GemmAmaxSm100(APIBase):
    def __init__(
        self,
        sample_a: torch.Tensor,
        sample_b: torch.Tensor,
        sample_sfa: torch.Tensor,
        sample_sfb: torch.Tensor,
        sample_c: torch.Tensor,
        sample_amax: torch.Tensor,
        acc_dtype: torch.dtype = torch.float32,
        mma_tiler_mn: Tuple[int, int] = (128, 128),
        cluster_shape_mn: Tuple[int, int] = (1, 1),
        sf_vec_size: int = 32,
    ):
        super().__init__()

        self._warn_experimental_api()
        self._logger.debug("Entering __init__")

        self.a_desc = self._make_tensor_desc(sample_a, name="sample_a")
        self.b_desc = self._make_tensor_desc(sample_b, name="sample_b")
        self.sfa_desc = self._make_tensor_desc(sample_sfa, name="sample_sfa")
        self.sfb_desc = self._make_tensor_desc(sample_sfb, name="sample_sfb")
        self.c_desc = self._make_tensor_desc(sample_c, name="sample_c")
        self.amax_desc = self._make_tensor_desc(sample_amax, name="sample_amax")

        self.acc_dtype = acc_dtype
        self.mma_tiler_mn = mma_tiler_mn
        self.cluster_shape_mn = cluster_shape_mn
        self.sf_vec_size = sf_vec_size

        self.num_cluster_overlap_margin = int(os.getenv("CUDNNFE_CLUSTER_OVERLAP_MARGIN", "0"))
        self._logger.debug(f"setting num_cluster_overlap_margin: {self.num_cluster_overlap_margin}")

        # used to reshape sfa/sfb tensors to atom layout
        self.atom_m = (32, 4)
        self.atom_k = 4

        self._interpret_uint8_as_fp4x2 = True
        self._logger.debug(
            f"__init__ completed with args: sample_a {self.a_desc.shape}, sample_b {self.b_desc.shape}, sample_sfa {self.sfa_desc.shape}, sample_sfb {self.sfb_desc.shape}, sample_c {self.c_desc.shape}, sample_amax {self.amax_desc.shape}, acc_dtype {acc_dtype}, mma_tiler_mn {mma_tiler_mn}, cluster_shape_mn {cluster_shape_mn}, sf_vec_size {sf_vec_size}"
        )

    def check_support(self) -> bool:
        self._logger.debug("Entering check_support")

        self._logger.debug("Checking dtypes and sf_vec_size")
        ab_dtype = self._check_dtype(
            self.a_desc,
            dtype=[torch.float4_e2m1fn_x2, torch.uint8, torch.float8_e5m2, torch.float8_e4m3fn],
            name="A",
        )
        self._check_dtype(
            self.b_desc,
            dtype=ab_dtype,
            name="B",
            extra_error_msg="A and B tensor dtypes must match",
        )
        if ab_dtype == torch.uint8:
            self._logger.warning("Uint8 ab_dtype will be interpreted as packed fp4, not as native uint8")

        self._value_error_if(
            self.sf_vec_size not in {16, 32},
            f"Unsupported sf_vec_size: received {self.sf_vec_size}, expected {{16, 32}}",
        )

        sf_dtype = self._check_dtype(
            self.sfa_desc,
            dtype=[torch.float8_e8m0fnu, torch.float8_e4m3fn, torch.int8],
            name="sfa",
        )
        self._check_dtype(
            self.sfb_desc,
            dtype=sf_dtype,
            name="sfb",
            extra_error_msg="sfa and sfb tensor dtypes must match",
        )
        if sf_dtype == torch.int8:
            self._logger.warning("Int8 sf_dtype will be interpreted as float8_e8m0fnu, not as native int8")

        self._value_error_if(
            sf_dtype == torch.float8_e4m3fn and self.sf_vec_size == 32,
            "Unsupported sf_dtype and sf_vec_size combination: float8_e4m3fn and 32 is not supported",
        )
        self._value_error_if(
            ab_dtype in {torch.float8_e5m2, torch.float8_e4m3fn} and self.sf_vec_size == 16,
            f"Unsupported ab_dtype and sf_vec_size combination: {{float8_e5m2, float8_e4m3fn}} and 16 is not supported",
        )

        c_dtype = self._check_dtype(
            self.c_desc,
            dtype=[torch.float32, torch.float16, torch.bfloat16, torch.float8_e5m2, torch.float8_e4m3fn, torch.float4_e2m1fn_x2, torch.uint8],
            name="C",
        )
        self._value_error_if(
            self._is_fp4x2(c_dtype) and not self._is_fp4x2(ab_dtype),
            f"Unsupported c_dtype and ab_dtype combination: fp4 c_dtype requires fp4 ab_dtype, got {ab_dtype}",
        )
        self._not_implemented_error_if(
            self._is_fp8(c_dtype) and self._is_fp8(ab_dtype),
            "Unsupported c_dtype and ab_dtype combination: fp8 ab_dtype and fp8 c_dtype (fails to launch)",
        )
        self._check_dtype(
            self.acc_dtype,
            dtype=torch.float32,
            name="Accumulator",
            extra_error_msg="Accumulator must be float32",
        )

        self.ab_dtype = ab_dtype
        self.c_dtype = c_dtype

        self._logger.debug("Checking tensor layout")
        m, k, l = self.a_desc.shape
        n, _, _ = self.b_desc.shape
        _, _, m_div_atom_m0_m1, _, sf_k_div_atom_k, _ = self.sfa_desc.shape
        _, _, n_div_atom_m0_m1, _, sf_k_div_atom_k, _ = self.sfb_desc.shape

        self._check_tensor_shape(self.a_desc, (m, k, l), "A")
        self._check_tensor_shape(self.b_desc, (n, k, l), "B")
        self._check_tensor_shape(self.c_desc, (m, n, l), "C")
        self._check_tensor_shape(
            self.sfa_desc,
            (self.atom_m[0], self.atom_m[1], m_div_atom_m0_m1, self.atom_k, sf_k_div_atom_k, l),
            "sfa",
        )
        self._check_tensor_shape(
            self.sfb_desc,
            (self.atom_m[0], self.atom_m[1], n_div_atom_m0_m1, self.atom_k, sf_k_div_atom_k, l),
            "sfb",
        )
        self.amax_desc = self._pad_tensor_to_ndim(self.amax_desc, 3, "amax")
        self._check_tensor_shape(self.amax_desc, (1, 1, 1), "amax")

        expected_m_div_atom = ceil_div(m, self.atom_m[0] * self.atom_m[1])
        expected_n_div_atom = ceil_div(n, self.atom_m[0] * self.atom_m[1])
        self._value_error_if(
            m_div_atom_m0_m1 != expected_m_div_atom,
            f"Input/Output shape mismatch: expected m_div_atom_m0_m1 (sfa.shape[2]) = {expected_m_div_atom}, got {m_div_atom_m0_m1}",
        )
        self._value_error_if(
            n_div_atom_m0_m1 != expected_n_div_atom,
            f"Input/Output shape mismatch: expected n_div_atom_m0_m1 (sfb.shape[2]) = {expected_n_div_atom}, got {n_div_atom_m0_m1}",
        )

        # Check tensor strides
        _ = self._check_tensor_stride(
            self.a_desc,
            stride=[(1, m, m * k), (k, 1, m * k)],
            name="A",
        )
        _ = self._check_tensor_stride(
            self.b_desc,
            stride=[(1, n, n * k), (k, 1, n * k)],
            name="B",
        )
        _ = self._check_tensor_stride(
            self.c_desc,
            stride=[(1, m, m * n), (n, 1, m * n)],
            name="C",
        )

        # Derive major mode from stride order
        self.a_major = "m" if self.a_desc.stride_order == (0, 1, 2) else "k"
        self.b_major = "n" if self.b_desc.stride_order == (0, 1, 2) else "k"
        self.c_major = "m" if self.c_desc.stride_order == (0, 1, 2) else "n"

        self._value_error_if(
            self._is_fp4x2(ab_dtype) and not (self.a_major == "k" and self.b_major == "k"),
            f"Unsupported A or B tensor stride: Float4 tensors require k-major layout for hardware efficiency, got {self.a_major} and {self.b_major}",
        )
        self._value_error_if(
            self._is_fp4x2(c_dtype) and self.c_major == "m",
            f"Unsupported C tensor stride: Float4 tensors require n-major layout for hardware efficiency, got {self.c_major}",
        )

        self._logger.debug("Checking mma tiler and cluster shape")
        self._value_error_if(
            self.mma_tiler_mn[0] not in [128, 256],
            f"Unsupported mma_tiler_mn[0]: expected {{128, 256}}, got {self.mma_tiler_mn[0]}",
        )
        self._value_error_if(
            self.mma_tiler_mn[1] not in [128, 256],
            f"Unsupported mma_tiler_mn[1]: expected {{128, 256}}, got {self.mma_tiler_mn[1]}",
        )
        self._not_implemented_error_if(
            self.mma_tiler_mn[0] == 256,
            "mma_tiler_mn[0] == 256 currently hangs",
        )
        self._value_error_if(
            self._is_fp4x2(self.ab_dtype) and self.mma_tiler_mn[1] == 256 and k <= 128,
            f"mma_tiler_mn (X, 256) requires k > 128 (packed x2), got {k}",
        )
        self._value_error_if(
            not (self.cluster_shape_mn[0] % (2 if self.mma_tiler_mn[0] == 256 else 1) == 0),
            "Illegal cluster shape",
        )
        self._not_implemented_error_if(
            self.mma_tiler_mn == (128, 256) and self.sf_vec_size == 16 and c_dtype in {torch.float32, torch.float16, torch.bfloat16},
            "mma_tiler_mn (128, 256), sf_vec_size 16, c_dtype {torch.float32, torch.float16, torch.bfloat16} fails to launch",
        )

        # Special cluster shape check for scale factor multicasts.
        # Due to limited size of scale factors, we can't multicast among more than 4 CTAs.
        self._value_error_if(
            not (
                self.cluster_shape_mn[0] <= 4
                and self.cluster_shape_mn[1] <= 4
                and self.cluster_shape_mn[0] > 0
                and self.cluster_shape_mn[1] > 0
                and is_power_of_2(self.cluster_shape_mn[0])
                and is_power_of_2(self.cluster_shape_mn[1])
            ),
            f"Invalid cluster shape: expected cluster_shape_mn values in {{1, 2, 4}}, got {self.cluster_shape_mn}",
        )

        self._logger.debug("Checking tensor alignment")

        def check_contigous_16B_alignment(dtype, is_mode0_major, tensor_shape):
            major_mode_idx = 0 if is_mode0_major else 1
            num_major_elements = tensor_shape[major_mode_idx]
            num_contiguous_elements = 16 * 8 // (_convert_to_cutlass_data_type(dtype).width)
            return num_major_elements % num_contiguous_elements == 0

        self._value_error_if(
            not (
                check_contigous_16B_alignment(ab_dtype, self.a_major == "m", (m, k, l))
                and check_contigous_16B_alignment(ab_dtype, self.b_major == "n", (n, k, l))
                and check_contigous_16B_alignment(c_dtype, self.c_major == "m", (m, n, l))
            ),
            "Unsupported tensor alignment: tensors must be 16B aligned",
        )

        self._logger.debug("Checking environment")
        self._runtime_error_if(not torch.cuda.is_available(), "CUDA is not available")
        device = torch.cuda.current_device()
        major, minor = torch.cuda.get_device_capability(device)
        compute_capability = major * 10 + minor
        self._runtime_error_if(
            compute_capability < 100,
            f"GemmAmax requires SM100+ compute capability, but found SM{compute_capability} on device {device}",
        )

        self._kernel = Sm100BlockScaledPersistentDenseGemmKernel

        self._is_supported = True
        self._logger.debug("check_support completed successfully")
        return True

    def compile(self) -> None:
        self._logger.debug("Entering compile")
        self._ensure_support_checked()
        if self._compiled_kernel is not None:
            self._logger.debug("Kernel already compiled; skipping recompilation")
            return

        gemm_amax = self._kernel(
            sf_vec_size=self.sf_vec_size,
            mma_tiler_mn=self.mma_tiler_mn,
            cluster_shape_mn=self.cluster_shape_mn,
        )
        hardware_info = cutlass.utils.HardwareInfo()
        max_active_clusters = hardware_info.get_max_active_clusters(self.cluster_shape_mn[0] * self.cluster_shape_mn[1])
        max_active_clusters -= self.num_cluster_overlap_margin
        self._value_error_if(
            max_active_clusters <= 0,
            "max_active_clusters must be > 0 after applying overlap margin; reduce CUDNNFE_CLUSTER_OVERLAP_MARGIN",
        )

        self._logger.debug("Compiling gemm_amax")
        a_cute = self._make_fake_cute_tensor_from_desc(self.a_desc, assumed_align=16)
        b_cute = self._make_fake_cute_tensor_from_desc(self.b_desc, assumed_align=16)
        sfa_cute = self._make_fake_cute_tensor_from_desc(self.sfa_desc, assumed_align=16)
        sfb_cute = self._make_fake_cute_tensor_from_desc(self.sfb_desc, assumed_align=16)
        c_cute = self._make_fake_cute_tensor_from_desc(self.c_desc, assumed_align=16)
        amax_cute = self._make_fake_cute_tensor_from_desc(self.amax_desc, assumed_align=16)
        fake_stream = make_fake_stream(use_tvm_ffi_env_stream=False)

        _compiled_kernel = cute.compile(
            gemm_amax,
            a_tensor=a_cute,
            b_tensor=b_cute,
            sfa_tensor=sfa_cute,
            sfb_tensor=sfb_cute,
            c_tensor=c_cute,
            amax_tensor=amax_cute,
            max_active_clusters=max_active_clusters,
            stream=fake_stream,
            options="--enable-tvm-ffi",
        )

        def tensor_api(
            a_tensor: torch.Tensor,
            b_tensor: torch.Tensor,
            sfa_tensor: torch.Tensor,
            sfb_tensor: torch.Tensor,
            c_tensor: torch.Tensor,
            amax_tensor: torch.Tensor,
            stream: cuda.CUstream,
        ):
            amax_tensor = self._pad_tensor_to_ndim(amax_tensor, 3, "amax")

            return _compiled_kernel(
                a_tensor,
                b_tensor,
                sfa_tensor,
                sfb_tensor,
                c_tensor,
                amax_tensor,
                stream,
            )

        self._compiled_kernel = tensor_api

        for attr_name in tuple(vars(self)):
            if attr_name.startswith("sample_"):
                setattr(self, attr_name, None)

        self._logger.debug("Kernel compiled successfully")

    def execute(
        self,
        a_tensor: torch.Tensor,
        b_tensor: torch.Tensor,
        sfa_tensor: torch.Tensor,
        sfb_tensor: torch.Tensor,
        c_tensor: torch.Tensor,
        amax_tensor: torch.Tensor,
        current_stream: Optional[cuda.CUstream] = None,
    ) -> None:
        self._logger.debug("Entering execute")
        current_stream = self._get_default_stream(current_stream)

        self._runtime_error_if(
            self._compiled_kernel is None,
            "GemmAmaxSm100 kernel not compiled; call compile() first",
        )
        self._logger.debug("Executing with compiled kernel")

        self._compiled_kernel(
            a_tensor=a_tensor,
            b_tensor=b_tensor,
            sfa_tensor=sfa_tensor,
            sfb_tensor=sfb_tensor,
            c_tensor=c_tensor,
            amax_tensor=amax_tensor,
            stream=current_stream,
        )
        self._logger.debug("Executed with compiled kernel successfully")


import logging

_logger = logging.getLogger(__name__)
_cache_of_GemmAmaxSm100Objects = {}


def gemm_amax_wrapper_sm100(
    a_tensor: torch.Tensor,
    b_tensor: torch.Tensor,
    sfa_tensor: torch.Tensor,
    sfb_tensor: torch.Tensor,
    c_major: str = "n",
    c_dtype: torch.dtype = torch.float32,
    acc_dtype: torch.dtype = torch.float32,
    mma_tiler_mn: Tuple[int, int] = (128, 128),
    cluster_shape_mn: Tuple[int, int] = (1, 1),
    sf_vec_size: int = 32,
    stream: Optional[cuda.CUstream] = None,
) -> TupleDict:

    _logger.debug("gemm_amax_wrapper_sm100: Creating empty output tensors c and amax")

    m, _, l = a_tensor.shape
    n, _, l = b_tensor.shape
    c_tensor = None
    if c_major == "m":
        c_tensor = torch.empty_strided((m, n, l), (1, m, m * n), dtype=c_dtype, device=a_tensor.device)
    elif c_major == "n":
        c_tensor = torch.empty_strided((m, n, l), (n, 1, m * n), dtype=c_dtype, device=a_tensor.device)
    else:
        raise ValueError(f"c_major must be either 'm' or 'n', got {c_major}")
    amax_tensor = torch.full((1, 1, 1), -float("inf"), device=a_tensor.device, dtype=torch.float32)

    cache_key = (
        a_tensor.shape,
        b_tensor.shape,
        sfa_tensor.shape,
        sfb_tensor.shape,
        a_tensor.dtype,
        b_tensor.dtype,
        sfa_tensor.dtype,
        sfb_tensor.dtype,
        a_tensor.stride(),
        b_tensor.stride(),
        sfa_tensor.stride(),
        sfb_tensor.stride(),
        c_major,
        c_dtype,
        acc_dtype,
        mma_tiler_mn,
        cluster_shape_mn,
        sf_vec_size,
    )
    if cache_key in _cache_of_GemmAmaxSm100Objects:
        _logger.debug("gemm_amax_wrapper_sm100: Using previously cached GemmAmaxSm100 object")
        gemm_amax = _cache_of_GemmAmaxSm100Objects[cache_key]
        gemm_amax.execute(
            a_tensor=a_tensor,
            b_tensor=b_tensor,
            sfa_tensor=sfa_tensor,
            sfb_tensor=sfb_tensor,
            c_tensor=c_tensor,
            amax_tensor=amax_tensor,
            current_stream=stream,
        )
    else:
        _logger.debug("gemm_amax_wrapper_sm100: No previously cached GemmAmaxSm100 object found, creating new GemmAmaxSm100 object")
        gemm_amax = GemmAmaxSm100(
            sample_a=a_tensor,
            sample_b=b_tensor,
            sample_sfa=sfa_tensor,
            sample_sfb=sfb_tensor,
            sample_c=c_tensor,
            sample_amax=amax_tensor,
            acc_dtype=acc_dtype,
            mma_tiler_mn=mma_tiler_mn,
            cluster_shape_mn=cluster_shape_mn,
            sf_vec_size=sf_vec_size,
        )
        assert gemm_amax.check_support()
        gemm_amax.compile()
        gemm_amax.execute(
            a_tensor=a_tensor,
            b_tensor=b_tensor,
            sfa_tensor=sfa_tensor,
            sfb_tensor=sfb_tensor,
            c_tensor=c_tensor,
            amax_tensor=amax_tensor,
            current_stream=stream,
        )
        _cache_of_GemmAmaxSm100Objects[cache_key] = gemm_amax

    return TupleDict(
        c_tensor=c_tensor,
        amax_tensor=amax_tensor,
    )
